#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Main module for connord"""

import argparse
import sys

# from .connord import update
from connord import update
from connord import version
from connord import listings
from connord import connect
from connord import iptables
from connord import user
from connord import servers
from connord import resources
from connord import areas


# pylint: disable=too-many-statements
def parse_args(argv):
    """Parse arguments
    :returns: list of args

    """
    description = """
Connect to NordVPN servers secure and fast.
DNS is managed with resolvconf and the firewall through iptables to keep
your connection safe.
"""
    parser = argparse.ArgumentParser(description=description)
    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument("-q", "--quiet", action="store_true", help="Be quiet")
    verbosity.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    command = parser.add_subparsers(dest="command")
    update_cmd = command.add_parser(
        "update", help="Update nordvpn configuration files."
    )
    update_cmd.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force update no matter of configuration.",
    )
    list_cmd = command.add_parser(
        "list", help="Prints all servers if no argument is given."
    )
    list_cmd.add_argument(
        "-c",
        "--country",
        action="append",
        nargs="?",
        help="select a specific country. may be specified multiple times. if \
                one of these arguments has no specifier then all country \
                codes are printed.",
    )
    list_cmd.add_argument(
        "-a",
        "--area",
        action="append",
        nargs="?",
        help="select a specific area.may be specified multiple times. if \
                one of these arguments has no specifier then all areas \
                are printed.",
    )
    list_cmd.add_argument(
        "-f",
        "--feature",
        action="append",
        nargs="?",
        help="select servers with a specific list of features. may be  \
                specified multiple times. if one of these arguments has no \
                specifier then all possible features are printed.",
    )
    list_cmd.add_argument(
        "-t",
        "--type",
        action="append",
        nargs="?",
        help="select servers with a specific type. may be specified multiple \
                times. if one of these arguments has no specifier then all \
                possible types are printed.",
    )
    list_cmd.add_argument(
        "--netflix", action="store_true", help="Select servers configured for netflix."
    )
    list_load_group = list_cmd.add_mutually_exclusive_group()
    list_load_group.add_argument(
        "--max-load", dest="max_load", type=int, help="Filter servers by maximum load."
    )
    list_load_group.add_argument(
        "--min-load", dest="min_load", type=int, help="Filter servers by minimum load."
    )
    list_load_group.add_argument(
        "--load", type=int, help="Filter servers by exact load match."
    )
    list_cmd.add_argument("--top", type=int, help="Show TOP count resulting servers.")
    list_cmd.add_argument(
        "--iptables", action="store_true", help="List all rules in iptables"
    )
    connect_cmd = command.add_parser("connect", help="Connect to a server.")
    server_best = connect_cmd.add_mutually_exclusive_group()
    server_best.add_argument(
        "-s",
        "--server",
        type=str,
        nargs=1,
        help="Connect to a specific server. Arguments -c, -a, -f, -t have no \
        effect.",
    )
    server_best.add_argument(
        "-b",
        "--best",
        action="store_true",
        help="Use best server depending on server load. When multiple servers \
        got the same load use the one with the best ping.",
    )
    connect_cmd.add_argument(
        "-c",
        "--country",
        action="append",
        nargs="?",
        help="Select a specific country. May be specified multiple times.",
    )
    connect_cmd.add_argument(
        "-a",
        "--area",
        action="append",
        nargs="?",
        help="Select a specific area. May be specified multiple times.",
    )
    connect_cmd.add_argument(
        "-f",
        "--feature",
        action="append",
        nargs="?",
        help="Select servers with a specific list of features. May be  \
                specified multiple times.",
    )
    connect_cmd.add_argument(
        "-t",
        "--type",
        action="append",
        nargs="?",
        help="Select servers with a specific type. May be specified multiple \
                times.",
    )
    connect_cmd.add_argument(
        "--netflix", action="store_true", help="Select servers configured for netflix."
    )
    connect_load_group = connect_cmd.add_mutually_exclusive_group()
    connect_load_group.add_argument(
        "--max-load", dest="max_load", type=int, help="Filter servers by maximum load."
    )
    connect_load_group.add_argument(
        "--min-load", dest="min_load", type=int, help="Filter servers by minimum load."
    )
    connect_load_group.add_argument(
        "--load", type=int, help="Filter servers by exact load match."
    )
    connect_cmd.add_argument(
        "-d", "--daemon", action="store_true", help="Start in daemon mode."
    )
    connect_cmd.add_argument(
        "-i",
        "--config",
        type=str,
        nargs="?",
        help="Take config from /path/to/config.{yml|ini}.",
    )
    connect_cmd.add_argument(
        "-o",
        "--openvpn",
        dest="openvpn_options",
        type=str,
        nargs=1,
        help="Options to pass through to openvpn as single string",
    )
    udp_tcp = connect_cmd.add_mutually_exclusive_group()
    udp_tcp.add_argument(
        "--udp", action="store_true", help="Use UDP protocol. This is the default"
    )
    udp_tcp.add_argument(
        "--tcp",
        action="store_true",
        help="Use TCP protocol. Only one of --udp or --tcp may be present.",
    )
    command.add_parser(
        "kill", help="Kill all processes of openvpn. Useful in daemon mode."
    )
    iptables_cmd = command.add_parser("iptables", help="Wrapper around iptables.")
    iptables_cmd_subparsers = iptables_cmd.add_subparsers(dest="iptables_sub")
    iptables_cmd_subparsers.add_parser("reload", help="Reload iptables")
    flush_cmd = iptables_cmd_subparsers.add_parser("flush", help="Flush iptables")
    flush_cmd.add_argument(
        "--no-fallback",
        dest="no_fallback",
        action="store_true",
        help="Flush tables ignoring fallback files",
    )
    apply_cmd = iptables_cmd_subparsers.add_parser(
        "apply", help="Apply iptables rules defined in configuration"
    )
    apply_cmd.add_argument(
        "domain", type=str, nargs=1, help="Apply iptables rules with domain"
    )
    udp_tcp = apply_cmd.add_mutually_exclusive_group()
    udp_tcp.add_argument(
        "--udp", action="store_true", help="Use UDP protocol. This is the default"
    )
    udp_tcp.add_argument(
        "--tcp",
        action="store_true",
        help="Use TCP protocol. Only one of --udp or --tcp may be present.",
    )
    command.add_parser("version", help="Show version")

    return parser.parse_args(argv)


def process_list_cmd(args):
    """
    Process arguments when command is 'list'

    :param args: Command-line arguments
    :returns: True if processing was successful
    """

    if args.iptables:
        return listings.list_iptables(["filter"], "4")

    _countries = args.country
    _area = args.area
    _types = args.type
    _features = args.feature
    _netflix = args.netflix
    _top = args.top

    if args.max_load:
        _load = args.max_load
        _match = "max"
    elif args.min_load:
        _load = args.min_load
        _match = "min"
    elif args.load:
        _load = args.load
        _match = "exact"
    else:  # apply defaults
        _load = 100
        _match = "max"

    return listings.main(
        _countries, _area, _types, _features, _netflix, _load, _match, _top
    )


def process_connect_cmd(args):
    """
    Process arguments for connect command
    :param object args: Command-line arguments
    :returns: True if processing was successful
    """

    if args.server:
        _server = args.server
    elif args.best:
        _server = "best"
    else:  # apply default
        _server = "best"

    _countries = args.country
    _areas = args.area
    _features = args.feature
    _types = args.type
    _netflix = args.netflix

    if args.max_load:
        _load = args.max_load
        _match = "max"
    elif args.min_load:
        _load = args.min_load
        _match = "min"
    elif args.load:
        _load = args.load
        _match = "exact"
    else:  # apply defaults
        _load = 10
        _match = "max"

    _daemon = args.daemon
    _config = args.config
    _openvpn = args.openvpn_options

    if args.udp:
        _protocol = "udp"
    elif args.tcp:
        _protocol = "tcp"
    else:  # apply default
        _protocol = "udp"

    return connect.connect(
        _server,
        _countries,
        _areas,
        _features,
        _types,
        _netflix,
        _load,
        _match,
        _daemon,
        _config,
        _openvpn,
        _protocol,
    )


@user.needs_root
def process_iptables_cmd(args):
    if args.iptables_sub == "flush":
        if args.no_fallback:
            iptables.reset(fallback=False)
        else:
            iptables.reset()
    elif args.iptables_sub == "apply":
        iptables.reset()
        if args.tcp:
            _protocol = "tcp"
        else:
            _protocol = "udp"

        domain = args.domain[0]
        _server = servers.get_server_by_domain(domain)
        if iptables.apply_config_dir(_server, _protocol):
            stats_dict = resources.get_stats()
            stats_dict["last_server"] = {}
            stats_dict["last_server"]["domain"] = domain
            stats_dict["last_server"]["protocol"] = _protocol
            resources.write_stats(stats_dict)
    elif args.iptables_sub == "reload":
        stats_dict = resources.get_stats()
        domain = str()
        _protocol = str()
        try:
            domain = stats_dict["last_server"]["domain"]
            _protocol = stats_dict["last_server"]["protocol"]
        except KeyError:
            print("Could not reload iptables. Apply iptables first.")
            return False

        _server = servers.get_server_by_domain(domain)

        iptables.reset()
        iptables.apply_config_dir(_server, _protocol)
    else:
        raise NotImplementedError("Not implemented")

    return True


def main():
    """Entry Point for the program.
    """

    if not sys.argv[1:]:
        sys.argv.extend(["-h"])

    args = parse_args(sys.argv[1:])
    try:
        if args.command == "update":
            update.update(force=args.force)
        elif args.command == "version":
            version.print_version()
        elif args.command == "list":
            process_list_cmd(args)
        elif args.command == "connect":
            process_connect_cmd(args)
        elif args.command == "kill":
            connect.kill_openvpn()
        elif args.command == "iptables":
            process_iptables_cmd(args)
        else:
            raise NotImplementedError("Could not process commandline arguments.")
    except PermissionError:
        print(
            'Permission Denied: You need to run "connord {}" as root'.format(
                args.command
            ),
            file=sys.stderr,
        )
        sys.exit(1)
    except resources.ResourceNotFoundError as error:
        print(error)
        sys.exit(1)
    except resources.MalformedResourceError as error:
        print(error)
        sys.exit(1)
    except areas.AreaError as error:
        print(error)
        sys.exit(1)
    except iptables.IptablesError as error:
        print(error)
        sys.exit(1)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sqlite3
from sqlite3 import Error
from cachetools import cached, TTLCache
from connord import ConnordError
from connord import resources


class SqliteError(ConnordError):
    """Thrown within this module"""


def create_connection(db_file=None):
    """Create a database connection"""
    if not db_file:
        db_file = resources.get_database_file()

    try:
        connection = sqlite3.connect(db_file)
        return connection
    except Error as error:
        print(error)

    return None


def create_table(connection, create_table_sql):
    try:
        cursor = connection.cursor()
        cursor.execute(create_table_sql)
    except Error as error:
        print(error)


def create_location(connection, location):
    sql = """ INSERT OR IGNORE INTO locations(
                latitude,
                longitude,
                display_name,
                city,
                country,
                country_code
              )
              VALUES(?,?,?,?,?,?) """

    try:
        cursor = connection.cursor()
        cursor.execute(sql, location)
        return cursor.lastrowid
    except Error as error:
        print(error)


def location_exists(connection, latitude, longitude):
    sql = """SELECT latitude, longitude FROM locations
            WHERE latitude = {} AND longitude = {}""".format(
        latitude, longitude
    )

    try:
        cursor = connection.cursor()
        result = cursor.execute(sql).fetchone()
        return bool(result)
    except Error as error:
        print(error)


def get_city(connection, latitude, longitude):
    sql = """SELECT city FROM locations
            WHERE latitude = {} AND longitude = {}""".format(
        latitude, longitude
    )

    try:
        cursor = connection.cursor()
        result = cursor.execute(sql).fetchone()
        return result
    except Error as error:
        print(error)


@cached(cache=TTLCache(ttl=60, maxsize=5))
def get_locations(connection):
    connection.row_factory = sqlite3.Row
    sql = """SELECT * FROM locations"""

    try:
        cursor = connection.cursor()
        result = cursor.execute(sql).fetchall()
        return result
    except Error as error:
        print(error)


def get_display_name(connection, latitude, longitude):
    sql = """SELECT display_name FROM locations
            WHERE latitude = {} AND longitude = {}""".format(
        latitude, longitude
    )

    try:
        cursor = connection.cursor()
        result = cursor.execute(sql).fetchone()
        return result
    except Error as error:
        print(error)


def create_location_table(connection):
    sql_create_location_table = """ CREATE TABLE IF NOT EXISTS locations(
                                        latitude text NOT NULL,
                                        longitude text NOT NULL,
                                        display_name text NOT NULL,
                                        city text NOT NULL,
                                        country NOT NULL,
                                        country_code NOT NULL,
                                        UNIQUE(latitude, longitude)
                                    ); """

    if connection:
        create_table(connection, sql_create_location_table)
    else:
        raise SqliteError("Could not create a database connection.")



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Everything around version
"""

from connord import __version__

__COPYRIGHT = """connord  Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
This program comes with ABSOLUTELY NO WARRANTY; This is free software, and you
are welcome to redistribute it under certain conditions; See the LICENSE file
shipped with this software for details."""


# TODO: Rename to show()
def print_version():
    """Prints the version of connord
    """
    # TODO: Show config in use along with the version
    print("connord", __version__)
    print(__COPYRIGHT)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import requests
from cachetools import cached, TTLCache
from connord import ConnordError
from connord.types import map_types_reverse
from connord import countries
from connord.formatter import Formatter

__API_URL = "https://api.nordvpn.com/server"
NETFLIX = ["us", "ca", "jp", "gb", "fr", "it"]


class ServerError(ConnordError):
    """Thrown within this module"""


def get_server_by_domain(domain):
    servers = get_servers()
    for server in servers:
        if server["domain"] == domain + ".nordvpn.com":
            return server

    return None


def get_servers_by_domains(domains):
    filtered_servers = [get_server_by_domain(domain) for domain in domains]

    return filtered_servers


@cached(cache=TTLCache(maxsize=1, ttl=60))
def get_servers():
    header = {
        "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:60.0) \
Gecko/20100101 Firefox/60.0"
    }

    with requests.get(__API_URL, headers=header, timeout=1) as response:
        return response.json()


def filter_netflix_servers(servers, _countries):
    servers = servers.copy()
    servers = countries.filter_servers(servers, NETFLIX)
    return servers


class ServersPrettyFormatter(Formatter):
    def format_headline(self, sep="="):
        headline = self.format_ruler(sep) + "\n"
        headline += "      {:25}  {:6}  {:15}  {:>9}  {}\n".format(
            "Country", "Domain", "IP Address", "Load", "Type"
        )
        headline += "      {}\n".format("Features")
        headline += self.format_ruler(sep)
        return headline

    def format_server(self, server, count, sep="-"):
        ident, _, _ = server["domain"].split(".")
        country = server["country"]
        ip = server["ip_address"]
        load = server["load"]

        categories = [category["name"] for category in server["categories"]]
        categories = ",".join(map_types_reverse(categories))

        features = [
            feature for feature in server["features"] if server["features"][feature]
        ]
        features = ",".join(features)

        string = "{:4d}: {:25}  {:6}  {:15}  load: {:>3d}  {}\n".format(
            count, country, ident, ip, load, categories
        )
        string += "      {}\n".format(features)
        string += self.format_ruler(sep)

        return string


def to_string(servers, stream=False):
    if not servers:
        return str()

    formatter = ServersPrettyFormatter()
    _file = formatter.get_stream_file(stream)

    headline = formatter.format_headline()
    print(headline, file=_file)
    count = 1
    for server in servers:
        _server_s = formatter.format_server(server, count)
        print(_server_s, file=_file)
        count += 1

    return formatter.get_output()



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import abc
from connord import ConnordError


class LoadError(ConnordError):
    """The Error thrown within this module"""


class Filter(metaclass=abc.ABCMeta):
    """
    Abstract class for different filters applied to servers.
    """

    __MAX = 100
    __MIN = 0

    def __init__(self, servers):
        """
        Initialise a Filter
        :param servers: A list of servers

        """
        self.servers = servers

    def __verify_load(self, load):
        """
        Verify if load is valid

        :load: Expects an integer between 0 and 100 inclusive
        :returns: True if load is valid
        :raises LoadError: If load is invalid
        """

        if not isinstance(load, int):
            raise LoadError('Wrong type "{}": {!s}'.format(type(load), load))
        elif load > self.__MAX or load < self.__MIN:
            raise LoadError("Load must be >= 0 and <= 100.")
        else:
            return True

    def apply(self, load):
        """
        Apply filters on servers

        :param load: Expects an integer between 0 and 100 inclusive
        :returns: Filtered list of servers
        """
        servers = self.servers

        if servers is None:
            raise TypeError("Servers may not be None")
        elif load is None or not servers:
            return servers
        else:
            self.__verify_load(load)

        filtered_servers = self._filter(load)

        return filtered_servers

    @abc.abstractmethod
    def _filter(self, load):
        """
        To be implemented by subclasses
        :param load: Expects an integer between 0 and 100 inclusive
        :returns: A list of filtered servers
        """


class LoadFilter(Filter):
    """
    Filter to match load exactly
    """

    def _filter(self, load):
        filtered_servers = [server for server in self.servers if server["load"] == load]
        return filtered_servers


class MaxLoadFilter(Filter):
    """
    Filter to match maximum load
    """

    def _filter(self, load):
        filtered_servers = [server for server in self.servers if server["load"] <= load]
        return filtered_servers


class MinLoadFilter(Filter):
    """
    Filter to match minimum load
    """

    def _filter(self, load):
        filtered_servers = [server for server in self.servers if server["load"] >= load]
        return filtered_servers


def filter_servers(servers, load, match="max"):
    """
    Filter a list of servers by load.

    :param servers: List of servers.
    :param load: An integer between 0 and 100 inclusive.
    :param match: Match load 'exact', 'max' or 'min'. Default is 'max'.
    :returns: The filtered list.
    :raises ValueError: If value of match is invalid.
    """

    servers = servers.copy()
    if match == "exact":
        return LoadFilter(servers).apply(load)
    elif match == "max":
        return MaxLoadFilter(servers).apply(load)
    elif match == "min":
        return MinLoadFilter(servers).apply(load)
    else:
        raise ValueError('Match must be one of "exact","max" or "min".')



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# here's a one liner to create this dictionary of country codes from
# nordvpn's api
# curl https://api.nordvpn.com/server 2>/dev/null | python -m json.tool | \
# grep 'flag\|country' | perl -ne 'print unless $seen{$_}++' | \
# sed -E -e 's/^        //' -e 's/("flag": )(".*")/\1\L\2/' -e 's/^"flag": //' \
# -e 's/^"country": //' | sed 's/,/:/;n' | paste - - -d' ' | tr '"' "'" | sort | \
# head -c -2 | sed -e '$a\'

from connord import ConnordError

COUNTRIES = {
    "ae": "United Arab Emirates",
    "al": "Albania",
    "ar": "Argentina",
    "at": "Austria",
    "au": "Australia",
    "ba": "Bosnia and Herzegovina",
    "be": "Belgium",
    "bg": "Bulgaria",
    "br": "Brazil",
    "ca": "Canada",
    "ch": "Switzerland",
    "cl": "Chile",
    "cr": "Costa Rica",
    "cy": "Cyprus",
    "cz": "Czech Republic",
    "de": "Germany",
    "dk": "Denmark",
    "ee": "Estonia",
    "eg": "Egypt",
    "es": "Spain",
    "fi": "Finland",
    "fr": "France",
    "gb": "United Kingdom",
    "ge": "Georgia",
    "gr": "Greece",
    "hk": "Hong Kong",
    "hr": "Croatia",
    "hu": "Hungary",
    "id": "Indonesia",
    "ie": "Ireland",
    "il": "Israel",
    "in": "India",
    "is": "Iceland",
    "it": "Italy",
    "jp": "Japan",
    "kr": "South Korea",
    "lu": "Luxembourg",
    "lv": "Latvia",
    "md": "Moldova",
    "mk": "North Macedonia",
    "mx": "Mexico",
    "my": "Malaysia",
    "nl": "Netherlands",
    "no": "Norway",
    "nz": "New Zealand",
    "pl": "Poland",
    "pt": "Portugal",
    "ro": "Romania",
    "rs": "Serbia",
    "se": "Sweden",
    "sg": "Singapore",
    "si": "Slovenia",
    "sk": "Slovakia",
    "th": "Thailand",
    "tr": "Turkey",
    "tw": "Taiwan",
    "ua": "Ukraine",
    "us": "United States",
    "vn": "Vietnam",
    "za": "South Africa",
}


class CountryError(ConnordError):
    """Thrown within this module"""


def verify_countries(countries):
    """Verify if a list of countries are valid

    :param countries: a list of countries to verify
    :returns: True if all countries are valid
    :raises CountryError: if one country in the list is invalid
    """

    if not isinstance(countries, list):
        raise CountryError("Wrong countries: {!s}".format(countries))

    wrong_countries = [
        country for country in countries if country.lower() not in COUNTRIES.keys()
    ]

    if wrong_countries:
        raise CountryError("Wrong countries: {!s}".format(wrong_countries))

    return True


# TODO: if countries are present twice this should filter countries to that
# specific country. I.e. when using the netflix option adding a country flag
# should narrow down the results instead of ignoring the double flag.
def filter_servers(servers, countries=None):
    """
    Filter a list of servers by country. If multiple countries are given
    match servers in country1 OR country2 OR ...

    :param servers: List of servers as parsed from nordvpn api in json format
    :param countries: List of countries to filter the server list. If None or
                     list is empty it returns all servers.
    :returns: The filtered list
    """

    if servers is None:
        raise TypeError("Servers may not be None")

    if countries is None or not countries or not servers:
        return servers

    countries_lower = [str.lower(country) for country in countries]
    verify_countries(countries_lower)

    # TODO: test if list comprehension is faster
    filtered_servers = []
    servers = servers.copy()
    for server in servers:
        flag = server["flag"].lower()
        for country in countries_lower:
            if country == flag:
                filtered_servers.append(server)
                break

    return filtered_servers


def to_string():
    """
    Assemble all countries to a printable string

    : returns: A simple formatted string for use as output on screen
    """
    result = ""
    for country_code, country in COUNTRIES.items():
        result += "{:6}{}\n".format(country_code, country)

    return result.rstrip()


# TODO: delete
def print_countries():
    """Prints all possible countries"""
    print(to_string())



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""List servers"""

from connord import servers
from connord import countries
from connord import types
from connord import features
from connord import load
from connord import iptables
from connord import areas
from connord.formatter import Formatter


def filter_servers_by_count(_servers, _top):
    """
    Filter servers to just show _top count results

    :param _servers: List of servers
    :param _top: Integer to show count servers
    :returns: The filtered servers
    """

    _servers = _servers.copy()
    filtered_servers = [server for i, server in enumerate(_servers, 1) if i <= _top]
    return filtered_servers


def list_iptables(tables, version):
    iptables.print_iptables(tables, version)
    return True


class OverviewPrettyFormatter(Formatter):
    def __init__(self, output=None, max_line_length=80, stream=False):
        super().__init__(output, max_line_length)
        self.stream = stream
        self.has_output = False

    def print_countries(self, _countries):
        if _countries and None in _countries:
            self.has_output = True
            _file = self.get_stream_file(self.stream)

            country_header = "Countries"
            print(self.center_string(country_header, sep="="), file=_file)
            for country_code, country in countries.COUNTRIES.items():
                print("  {:6}{}".format(country_code, country), file=_file)
            print(self.format_ruler(sep="-"), file=_file)

    def print_types(self, _types):
        if _types and None in _types:
            self.has_output = True
            _file = self.get_stream_file(self.stream)

            types_header = self.center_string("Server Types", sep="=")
            print(types_header, file=_file)
            for server_type, description in types.TYPES.items():
                print("  {:26}{}".format(server_type, description), file=_file)
            print(self.format_ruler(sep="-"), file=_file)

    def print_features(self, _features):
        if _features and None in _features:
            self.has_output = True
            _file = self.get_stream_file(self.stream)

            features_header = "Server Features"
            print(self.center_string(features_header, sep="="), file=_file)
            for feature, description in features.FEATURES.items():
                print("  {:26}{}".format(feature, description), file=_file)
            print(self.format_ruler(sep="-"), file=_file)

    def print_areas(self, _areas):
        if _areas and None in _areas:
            self.has_output = True
            areas.print_areas()


def filter_servers(
    _servers, _netflix, _countries, _areas, _features, _types, _load, _match, _top
):
    _servers = _servers.copy()
    if _load:
        _servers = load.filter_servers(_servers, _load, _match)
    if _netflix:
        _servers = servers.filter_netflix_servers(_servers, _countries)
    if _countries:
        _servers = countries.filter_servers(_servers, _countries)
    if _areas:
        _servers = areas.filter_servers(_servers, _areas)
    if _types:
        _servers = types.filter_servers(_servers, _types)
    if _features:
        _servers = features.filter_servers(_servers, _features)
    if _top:
        _servers = filter_servers_by_count(_servers, _top)

    return _servers


# TODO: rename to list_servers
def main(_countries, _areas, _types, _features, _netflix, _load, _match, _top):
    """
    Main method to do the actual listing and prints the resulting
    list of servers.

    :param _countries: List of countries
    :param _area: List of areas
    :param _types: List of types
    :param _features: List of features
    :param _netflix: If set filter servers optimised for netflix
    :param _load: An integer to filter servers by load.
    :param _match: Apply load filter with 'max', 'min' or 'exact' match
    :param _top: Show just _top count results.
    :returns: True
    """

    formatter = OverviewPrettyFormatter(stream=True)
    formatter.print_countries(_countries)
    formatter.print_areas(_areas)
    formatter.print_types(_types)
    formatter.print_features(_features)

    if not formatter.has_output:
        _servers = servers.get_servers()
        _servers = filter_servers(
            _servers,
            _netflix,
            _countries,
            _areas,
            _features,
            _types,
            _load,
            _match,
            _top,
        )

        servers.to_string(_servers, stream=True)

    return True



"""setup.py"""
from setuptools import setup, find_packages
from connord import __version__


def readme():
    """Parse README.md

    : returns: README.md as string
    """
    with open("README.md") as file_handle:
        return file_handle.read()


setup(
    name="connord",
    version=__version__,
    description="Connect to NordVPN servers secure and fast",
    long_description=readme(),
    long_description_content_type="text/markdown; charset=UTF-8; variant=GFM",
    keywords="nordvpn vpn secure privacy openvpn iptables",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "Topic :: Security",
        "Topic :: System :: Networking",
        "Topic :: System :: Networking :: Firewalls",
        "Topic :: System :: Systems Administration",
    ],
    install_requires=[
        "requests>=2.22.0,<3.0.0",
        "cachetools>=3.1.1,<4.0.0",
        "Jinja2>=2.10.1,<3.0.0",
        "netaddr>=0.7.19,<1.0.0",
        "PyYAML>=5.1,<6.0",
        "python-iptables>=0.14.0,<1.0.0",
        "progress>=1.5,<2.0",
    ],
    tests_require=["pytest", "pytest-cov", "vermin", "mock"],
    python_requires=">=3.6",
    packages=find_packages(),
    include_package_data=True,
    entry_points={"console_scripts": ["connord = connord.__main__:main"]},
    url="http://gitlab.noiams.space/maelstor/connord",
    author="Mael Stor",
    author_email="maelstor@posteo.de",
    license="GPL-3.0-or-later",
    zip_safe=False,
)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Update configuration files of nordvpn
"""
# TODO: improve exception handling

import os
from shutil import move
from zipfile import ZipFile
from datetime import datetime, timedelta
import requests
from connord import ConnordError
from connord import resources
from connord import user

__URL = "https://downloads.nordcdn.com/configs/archives/servers/ovpn.zip"
TIMEOUT = timedelta(days=1)


class UpdateError(ConnordError):
    """Raised during update"""


@user.needs_root
def update_orig():
    """
    Move the original file to make room for the newly downloaded file
    """

    zip_file = resources.get_zip_file(create_dirs=True)
    move(zip_file, zip_file + ".orig")


def get():
    """Get the zip file
    """
    zip_path = resources.get_zip_path()
    update_orig()

    print("Downloading {} ...".format(zip_path))
    with requests.get(__URL, stream=True, timeout=1) as response, open(
        zip_path, "wb"
    ) as zip_fd:
        for chunk in response.iter_content(chunk_size=512):
            zip_fd.write(chunk)

    return True


def file_equals(_file, _other):
    """Compares the orig.zip file to the downloaded file
    : returns: False if file sizes differ
    """
    if os.path.exists(_file) and os.path.exists(_other):
        return os.path.getsize(_file) == os.path.getsize(_other)
    else:
        return False


def unzip():
    """Unzip the configuration files
    """
    zip_dir = resources.get_zip_dir()
    zip_file = resources.get_zip_file()
    print("Unzipping {} ...".format(zip_file))
    with ZipFile(zip_file) as zip_stream:
        zip_stream.extractall(zip_dir)


def update(force=False):
    """Update the nordvpn configuration files
    """
    if force:
        get()
        unzip()
    else:
        zip_file = resources.get_zip_file()
        orig_file = resources.get_zip_file("ovpn.zip.orig")
        if update_needed():
            get()
            if not file_equals(orig_file, zip_file):
                unzip()
            else:
                print(zip_file + " already up-to-date")
        else:
            next_update = datetime.fromtimestamp(os.path.getctime(zip_file)) + TIMEOUT
            print("No update needed. Next update needed at {!s}".format(next_update))

    return True


def update_needed():
    """Check if an update is needed
    : returns: False if the zip file's creation time hasn't reached the timeout
               else True.
    """
    try:
        zip_file = resources.get_zip_file()
        now = datetime.now()
        time_created = datetime.fromtimestamp(os.path.getctime(zip_file))
        return now - TIMEOUT > time_created
    except resources.ResourceNotFoundError:
        return True
